"""Matcher helpers."""
from re import compile as compile_re
from typing import cast, Any, Optional, Pattern, Type


__all__ = [
    "Anything",
    "IsA",
    "IsNot",
    "StringMatching",
    "ErrorMatching",
]


class _Anything:
    def __eq__(self, target: object) -> bool:
        """Return true if target is not None."""
        return target is not None

    def __repr__(self) -> str:
        """Return a string representation of the matcher."""
        return "<Anything>"


def Anything() -> Any:
    """Match anything except None.

    Example:
        ```python
        assert "foobar" == Anything()
        assert None != Anything()
        ```
    """
    return _Anything()


class _IsA:
    _match_type: type

    def __init__(self, match_type: type) -> None:
        """Initialize the matcher with a type."""
        self._match_type = match_type

    def __eq__(self, target: object) -> bool:
        """Return true if target is a self._match_type."""
        return type(target) == self._match_type

    def __repr__(self) -> str:
        """Return a string representation of the matcher."""
        return "<IsA {self._match_type.__name__}>"


def IsA(match_type: type) -> Any:
    """Match anything that satisfies the passed in type.

    Arguments:
        match_type: Type to match.

    Example:
        ```python
        assert "foobar" == IsA(str)
        assert datetime.now() == IsA(datetime)
        assert 42 == IsA(int)
        ```
    """
    return _IsA(match_type)


class _IsNot:
    _reject_value: object

    def __init__(self, value: object) -> None:
        """Initialize the matcher with the value to reject."""
        self._reject_value = value

    def __eq__(self, target: object) -> bool:
        """Return true if target is not self._reject_value."""
        return target != self._reject_value

    def __repr__(self) -> str:
        """Return a string representation of the matcher."""
        return f"<IsNot {self._reject_value}>"


def IsNot(value: object) -> Any:
    """Match anything that isn't the passed in value.

    Arguments:
        value: Value to check against.

    Example:
        ```python
        assert "foobar" == IsNot("bazquux")
        assert 42 == IsNot("the question")
        assert 1 != IsNot(1)
        ```
    """
    return _IsNot(value)


class _StringMatching:
    _pattern: Pattern[str]

    def __init__(self, match: str) -> None:
        """Initialize the matcher with the pattern to match."""
        self._pattern = compile_re(match)

    def __eq__(self, target: object) -> bool:
        """Return true if target is not self._reject_value."""
        return (
            type(target) is str and self._pattern.search(cast(str, target)) is not None
        )

    def __repr__(self) -> str:
        """Return a string representation of the matcher."""
        return f"<StringMatching {self._pattern.pattern}>"


def StringMatching(match: str) -> str:
    """Match any string matching the passed in pattern.

    Arguments:
        match: Pattern to check against; will be compiled into an re.Pattern.

    Example:
        ```python
        assert "foobar" == StringMatching("bar")
        assert "foobar" != StringMatching("^bar")
        ```
    """
    return cast(str, _StringMatching(match))


class _ErrorMatching:
    _error_type: Type[Exception]
    _string_matcher: Optional[_StringMatching]

    def __init__(self, error: Type[Exception], match: Optional[str] = None) -> None:
        """Initialize with the Exception type and optional message matcher."""
        self._error_type = error
        self._string_matcher = _StringMatching(match) if match is not None else None

    def __eq__(self, target: object) -> bool:
        """Return true if target is not self._reject_value."""
        error_match = type(target) == self._error_type
        message_match = (
            str(target) == self._string_matcher
            if self._string_matcher is not None
            else True
        )

        return error_match and message_match

    def __repr__(self) -> str:
        """Return a string representation of the matcher."""
        return (
            f"<ErrorMatching {self._error_type.__name__} match={self._string_matcher}>"
        )


def ErrorMatching(error: Type[Exception], match: Optional[str] = None) -> Exception:
    """Match any error matching an Exception type and optional message matcher.

    Arguments:
        error: Exception type to match against.
        match: Pattern to check against; will be compiled into an re.Pattern.

    Example:
        ```python
        assert ValueError("oh no!") == ErrorMatching(ValueError)
        assert ValueError("oh no!") == ErrorMatching(ValueError, match="no")
        ```
    """
    return cast(Exception, _ErrorMatching(error, match))



"""Decoy and stub configuration registry."""
from mock import Mock
from weakref import finalize, WeakValueDictionary
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from .mock import DecoyMock
from .stub import Stub
from .types import Call

if TYPE_CHECKING:
    DecoyMapType = WeakValueDictionary[int, DecoyMock]


class Registry:
    """Decoy and stub configuration registry.

    The Registry collects weak-references to decoys created in order to
    automatically clean up stub configurations when a decoy goes out of scope.
    """

    def __init__(self) -> None:
        """Initialize a Registry."""
        self._decoy_map: DecoyMapType = WeakValueDictionary()
        self._stub_map: Dict[int, List[Stub[Any]]] = {}

    def register_decoy(self, decoy: DecoyMock) -> int:
        """Register a decoy for tracking."""
        decoy_id = id(decoy)

        self._decoy_map[decoy_id] = decoy
        finalize(decoy, self._clear_decoy_stubs, decoy_id)

        return decoy_id

    def register_stub(self, decoy_id: int, stub: Stub[Any]) -> None:
        """Register a stub for tracking."""
        stub_list = self.get_decoy_stubs(decoy_id)
        self._stub_map[decoy_id] = stub_list + [stub]

    def get_decoy(self, decoy_id: int) -> Optional[Mock]:
        """Get a decoy by identifier."""
        return self._decoy_map.get(decoy_id)

    def get_decoy_stubs(self, decoy_id: int) -> List[Stub[Any]]:
        """Get a decoy's stub list by identifier."""
        return self._stub_map.get(decoy_id, [])

    def peek_decoy_last_call(self, decoy_id: int) -> Optional[Call]:
        """Get a decoy's last call."""
        decoy = self._decoy_map.get(decoy_id, None)

        if decoy is not None and len(decoy.mock_calls) > 0:
            return decoy.mock_calls[-1]

        return None

    def pop_decoy_last_call(self, decoy_id: int) -> Optional[Call]:
        """Pop a decoy's last call off of its call stack."""
        decoy = self._decoy_map.get(decoy_id, None)

        if decoy is not None and len(decoy.mock_calls) > 0:
            return decoy.mock_calls.pop()

        return None

    def _clear_decoy_stubs(self, decoy_id: int) -> None:
        if decoy_id in self._stub_map:
            del self._stub_map[decoy_id]



"""Decoy test double stubbing and verification library."""
from typing import cast, Any, Callable, Mapping, Optional, Sequence, Tuple, Type

from .mock import create_decoy_mock, DecoyMock
from .registry import Registry
from .stub import Stub
from .types import Call, ClassT, FuncT, ReturnT


class Decoy:
    """Decoy test double state container."""

    _registry: Registry
    _last_decoy_id: Optional[int]

    def __init__(self) -> None:
        """Initialize the state container for test doubles and stubs.

        You should initialize a new Decoy instance for every test.

        Example:
            ```python
            import pytest
            from decoy import Decoy

            @pytest.fixture
            def decoy() -> Decoy:
                return Decoy()
            ```
        """
        self._registry = Registry()
        self._last_decoy_id = None

    def create_decoy(self, spec: Type[ClassT], *, is_async: bool = False) -> ClassT:
        """Create a class decoy for `spec`.

        Arguments:
            spec: A class definition that the decoy should mirror.
            is_async: Set to `True` if the class has `await`able methods.

        Returns:
            A `MagicMock` or `AsyncMock`, typecast as an instance of `spec`.

        Example:
            ```python
            def test_get_something(decoy: Decoy):
                db = decoy.create_decoy(spec=Database)
                # ...
            ```

        """
        decoy = self._create_and_register_mock(spec=spec, is_async=is_async)
        return cast(ClassT, decoy)

    def create_decoy_func(
        self, spec: Optional[FuncT] = None, *, is_async: bool = False
    ) -> FuncT:
        """Create a function decoy for `spec`.

        Arguments:
            spec: A function that the decoy should mirror.
            is_async: Set to `True` if the function is `await`able.

        Returns:
            A `MagicMock` or `AsyncMock`, typecast as the function given for `spec`.

        Example:
            ```python
            def test_create_something(decoy: Decoy):
                gen_id = decoy.create_decoy_func(spec=generate_unique_id)
                # ...
            ```
        """
        decoy = self._create_and_register_mock(spec=spec, is_async=is_async)

        return cast(FuncT, decoy)

    def when(self, _rehearsal_result: ReturnT) -> Stub[ReturnT]:
        """Create a [Stub][decoy.stub.Stub] configuration using a rehearsal call.

        See [stubbing](/#stubbing) for more details.

        Arguments:
            _rehearsal_result: The return value of a rehearsal, used for typechecking.

        Returns:
            A Stub to configure using `then_return` or `then_raise`.

        Example:
            ```python
            db = decoy.create_decoy(spec=Database)
            decoy.when(db.exists("some-id")).then_return(True)
            ```
        """
        decoy_id, rehearsal = self._pop_last_rehearsal()
        stub = Stub[ReturnT](rehearsal=rehearsal)

        self._registry.register_stub(decoy_id, stub)

        return stub

    def verify(self, _rehearsal_result: Optional[ReturnT] = None) -> None:
        """Verify a decoy was called using a rehearsal.

        See [verification](/#verification) for more details.

        Arguments:
            _rehearsal_result: The return value of a rehearsal, unused.

        Example:
            ```python
            def test_create_something(decoy: Decoy):
                gen_id = decoy.create_decoy_func(spec=generate_unique_id)

                # ...

                decoy.verify(gen_id("model-prefix_"))
            ```
        """
        decoy_id, rehearsal = self._pop_last_rehearsal()
        decoy = self._registry.get_decoy(decoy_id)

        if decoy is None:
            raise ValueError("verify must be called with a decoy rehearsal")

        decoy.assert_has_calls([rehearsal])

    def _create_and_register_mock(self, spec: Any, is_async: bool) -> DecoyMock:
        decoy = create_decoy_mock(is_async=is_async, spec=spec)
        decoy_id = self._registry.register_decoy(decoy)
        side_effect = self._create_track_call_and_act(decoy_id)

        decoy.configure_mock(side_effect=side_effect)

        return decoy

    def _pop_last_rehearsal(self) -> Tuple[int, Call]:
        decoy_id = self._last_decoy_id

        if decoy_id is not None:
            rehearsal = self._registry.pop_decoy_last_call(decoy_id)
            self._last_decoy_id = None

            if rehearsal is not None:
                return (decoy_id, rehearsal)

        raise ValueError("when/verify must be called with a decoy rehearsal")

    def _create_track_call_and_act(self, decoy_id: int) -> Callable[..., Any]:
        def track_call_and_act(
            *args: Sequence[Any], **_kwargs: Mapping[str, Any]
        ) -> Any:
            self._last_decoy_id = decoy_id

            last_call = self._registry.peek_decoy_last_call(decoy_id)
            stubs = reversed(self._registry.get_decoy_stubs(decoy_id))

            if last_call is not None:
                for stub in stubs:
                    if stub._rehearsal == last_call:
                        return stub._act()

            return None

        return track_call_and_act



"""Custom unittest.mock subclasses."""

from mock import AsyncMock, MagicMock
from typing import Any, Union


class SyncDecoyMock(MagicMock):
    """MagicMock variant where all child mocks use the parent side_effect."""

    def _get_child_mock(self, **kwargs: Any) -> Any:
        return super()._get_child_mock(**kwargs, side_effect=self.side_effect)


class AsyncDecoyMock(AsyncMock):  # type: ignore[misc]
    """AsyncMock variant where all child mocks use the parent side_effect."""

    def _get_child_mock(self, **kwargs: Any) -> Any:
        return super()._get_child_mock(**kwargs, side_effect=self.side_effect)


DecoyMock = Union[SyncDecoyMock, AsyncDecoyMock]


def create_decoy_mock(is_async: bool, **kwargs: Any) -> DecoyMock:
    """Create a MagicMock or AsyncMock."""
    if is_async is False:
        return SyncDecoyMock(**kwargs)
    else:
        return AsyncDecoyMock(**kwargs)
